#include <stdio.h>
#include <stdlib.h>
#include "UDPstream.hh"

#ifdef WIN32
#include <windows.h>
#define SOCKERR(x) (x==SOCKET_ERROR)
#define ENOBUFS WSAENOBUFS
#else
#include <errno.h>
#include <sys/types.h>
#include <sys/socket.h>
#define SOCKERR(x) (x<0)
#endif

// breaks the buffer up into separate packets

UDPstream::UDPstream(int pt):
  RawUDP(pt),mtu(0),packetsize(576),actualbyterate(0.0),shortdelay(0.002){
  setByteRate(fMega(50.0));
}

UDPstream::UDPstream(char *host,int pt):
  RawUDP(host,pt),mtu(0),packetsize(576),actualbyterate(0.0),shortdelay(0.002){
  setByteRate(fMega(50.0));
}

/*
  Eventually need to tag the packets with a sequence number
  Can tag in-situ or always copy to an auxiliary buffer
 */
int UDPstream::send(char *buffer,int bufsize,int flgs){
  int r,n=0,sendsize;
  // how to sendto???
  if(flgs>=0) flags=flgs;
  // will fail if not bound or connected
  // OK, so now we break it into blocks
  packetdelay.reset();
  sendsize=(this->packetsize>bufsize)?bufsize:this->packetsize;
  r = ::send(port,buffer,sendsize,flags);
  if(SOCKERR(r)){
    if(errno==ENOBUFS){
      // try again after delay
      do{
	shortdelay.instantaneousWait();
	r = ::send(port,buffer,sendsize,flags);
      }while(SOCKERR(r) && errno==ENOBUFS);
      n+=r;
    }
    else {
      puts("Aborted with error which is not ENOBUFS");
      perror("error1 was ");
    }
  }
  else n+=r;
  // now do the rest ======================
  for(int i=sendsize;r>0 && i<bufsize;i+=this->packetsize,n+=r){
    int remaining = bufsize-i;
    sendsize=(this->packetsize>remaining)?remaining:this->packetsize;
    packetdelay.errorDiffusionWait();
    r = ::send(port,buffer,sendsize,flags);
    if(SOCKERR(r)){
      if(errno==ENOBUFS){ // try again after small delay
	do{
	  shortdelay.instantaneousWait();
	  r = ::send(port,buffer,sendsize,flags);
	}while(SOCKERR(r) && errno==ENOBUFS);
	n+=r;
      }
      else {
	puts("Aborted with error which is not ENOBUFS");
	perror("error2 was ");
      }
    }
    else n+=r;
    // manage ENOBUFS errors and fail on other errors
    // just do a short delay
  }
  return n;
}

int UDPstream::send(char *buffer,int bufsize,
		    IPaddress &to,int flgs){ // how to sendto???
  int r,n=0,sendsize;
  // how to sendto???
  if(flgs>=0) flags=flgs;
  // will fail if not bound or connected
  // OK, so now we break it into blocks
  packetdelay.reset();
  sendsize=(this->packetsize>bufsize)?bufsize:this->packetsize;
  r = ::sendto(port,buffer,sendsize,flags,
	       to.getSocketAddress(),sizeof(struct sockaddr_in));
  if(SOCKERR(r)){
    if(errno==ENOBUFS){
      // try again after delay
      do{
	shortdelay.instantaneousWait();
	r = ::sendto(port,buffer,sendsize,flags,
		     to.getSocketAddress(),sizeof(struct sockaddr_in));
      }while(SOCKERR(r) && errno==ENOBUFS);
    }
  }
  // now do the rest ======================
  for(int i=sendsize;r>0 && i<bufsize;i+=this->packetsize,n+=r){
    int remaining = bufsize-i;
    sendsize=(this->packetsize>remaining)?remaining:this->packetsize;
    packetdelay.errorDiffusionWait();
    r = ::sendto(port,buffer,sendsize,flags,
		 to.getSocketAddress(),sizeof(struct sockaddr_in));
    if(SOCKERR(r)){
      // manage ENOBUFS errors and fail on other errors
      // just do a short delay
      if(errno==ENOBUFS){ // try again after small delay
	do {
	  shortdelay.instantaneousWait();
	  r = ::sendto(port,buffer,sendsize,flags,
		       to.getSocketAddress(),sizeof(struct sockaddr_in));
	} while(SOCKERR(r) && errno==ENOBUFS);
      }
    }
  }
  return n;
}
int UDPstream::recv(char *buffer,int bufsize,IPaddress &from,int flgs){
  int n=0;
  for(int i=0;i<bufsize;i+=packetsize){
    int r,recvsize=bufsize-i;
    if(recvsize>packetsize) 
      recvsize=packetsize;
    r=RawUDP::recv(buffer,recvsize,from,flgs);
    if(!SOCKERR(r)) n+=r;
  }
  return n;
}
int UDPstream::recv(char *buffer,int bufsize,int flgs){
  int n=0;
  for(int i=0;i<bufsize;i+=packetsize){
    int r,recvsize=bufsize-i;
    if(recvsize>packetsize) 
      recvsize=packetsize;
    r=RawUDP::recv(buffer,recvsize,flgs);
    if(!SOCKERR(r)) n+=r;
  }
  return n;
}
int UDPstream::getMTU(){
  // how does MTU discovery work in this context???
  return 1440;
}

void UDPstream::setPacketRate(double packetspersecond){
  // implies measurement from packet-to-packet
  packetrate=packetspersecond;
  setPacketDelay(1.0/packetrate);
}

void UDPstream::setByteRate(double bytespersecond){
  this->setPacketRate(bytespersecond/((double)packetsize));
}

void UDPstream::setPacketDelay(double seconds){ // converts to integer internally
  packetdelay.setDelay(seconds);
  // make packetrate correct?
  if(seconds>0.0)
    packetrate=1.0/seconds;
  else // ALL BETS ARE OFF IF YOU MAKE THE DELAY 0!!!
    seconds=1e-6; // approx minimum measurable delay
}
// now we allow setting of timing requirments
void UDPstream::setPacketSize(long nbytes){
  double currentbyterate = this->getByteRate();
  this->packetsize = nbytes;
  this->setByteRate(currentbyterate);
}
